#External:
import netCDF4
import time
import os
from socket import error as SocketError
import errno
import warnings
import requests
import requests_cache
import datetime

#Internal:
import safe_handling
import requests_sessions

class Dataset:
    def __init__(self,url_name,remote_data_node='',timeout=120,cache=None,expire_after=datetime.timedelta(hours=1),session=None):
        self.url_name=url_name
        self.timeout=timeout
        self.cache=cache
        self.expire_after=expire_after
        self.passed_session=session
        return

    def __enter__(self):
        if (isinstance(self.passed_session,requests.Session) or
            isinstance(self.passed_session,requests_cache.core.CachedSession)
            ):
            self.session=self.passed_session
        else:
            self.session=requests_sessions.create_single_session(cache=self.cache,expire_after=self.expire_after)

        #Disable cache for streaming get:
        if isinstance(self.session,requests_cache.core.CachedSession):
            with self.session.cache_disabled():
                return self._initiate_query()
        else:
            return self._initiate_query()

    def _initiate_query(self):
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', message='Unverified HTTPS request is being made. Adding certificate verification is strongly advised. See: https://urllib3.readthedocs.org/en/latest/security.html')
            headers = {'connection': 'close'}
            self.response = self.session.get(self.url_name, 
                        cert=(os.environ['X509_USER_PROXY'],os.environ['X509_USER_PROXY']),
                        verify=False,
                        headers=headers,
                        allow_redirects=True,
                        timeout=self.timeout,
                        stream=True)

        if self.response.ok:
            try:
                #content size must be larger than 0.
                #when content-length key exists
                content_size=int(self.response.headers['Content-Length'])
                if content_size==0:
                   raise RemoteEmptyError('URL {0} is empty. It will not be considered'.format(self.url_name))
            except KeyError:
                #Assume success:
                pass
        return self

    def __exit__(self,type,value,traceback):
        self.response.close()
        if not (isinstance(self.passed_session,requests.Session) or
            isinstance(self.passed_session,requests_cache.core.CachedSession)
            ):
            self.session.close()
        return

class RemoteEmptyError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)



#External:
import netCDF4
import time
import os
from socket import error as SocketError
import errno
import warnings
import requests
import requests_cache
import datetime
import hashlib

#Internal:
import safe_handling
import esgf_http

class http_netCDF:
    def __init__(self,url_name,semaphores=dict(),remote_data_node='',timeout=120,cache=None,expire_after=datetime.timedelta(hours=1),session=None):
        self.url_name=url_name
        self.semaphores=semaphores
        self.timeout=timeout
        self.cache=cache
        self.expire_after=expire_after
        self.session=session

        if (remote_data_node in  self.semaphores.keys()):
            self.semaphore=semaphores[remote_data_node]
            self.handle_safely=True
        else:
            self.semaphore=safe_handling.dummy_semaphore()
            self.handle_safely=False
        return

    def __enter__(self):
        self.semaphore.acquire()
        return self

    def __exit__(self,type,value,traceback):
        if self.handle_safely:
            #Do not release semaphore right away if data is not local:
            time.sleep(0.01)
        self.semaphore.release()
        return

    def check_if_opens(self,num_trials=5):
        #If ftp, assume available:
        if len(self.url_name)>3 and self.url_name[:3]=='ftp':
            return True
        success=False
        for trial in range(num_trials):
            if not success:
                try:
                    with esgf_http.Dataset(self.url_name,
                                            cache=self.cache,
                                            timeout=self.timeout,
                                            expire_after=self.expire_after,
                                            session=self.session) as dataset:
                        pass
                    success=True
                except SocketError as e:
                    #http://stackoverflow.com/questions/20568216/python-handling-socket-error-errno-104-connection-reset-by-peer
                    if e.errno != errno.ECONNRESET:
                        raise
                    time.sleep(3*(trial+1))
                    pass
                except requests.exceptions.ReadTimeout as e:
                    time.sleep(3*(trial+1))
                    pass
                except esgf_http.RemoteEmptyError as e:
                    print(e)
                    break
        return success

    def check_if_opens_wget(self,num_trials=5):
        #If ftp, assume available:
        if len(url_name)>3 and url_name[:3]=='ftp':
            return True

        success=False
        for trial in range(num_trials):
            if not success:
                wget_call='wget --timeout={2} --spider --ca-directory={0} --certificate={1} --private-key={1}'.format(os.environ['X509_CERT_DIR'],os.environ['X509_USER_PROXY'],self.timeout).split(' ')
                wget_call.append(self.url_name)

                proc=subprocess.Popen(wget_call,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
                (out, err) = proc.communicate()

                status_string='HTTP request sent, awaiting response... '
                error_codes=[ int(line.replace(status_string,'').split(' ')[0]) for line in err.splitlines() if status_string in line]
                length_string='Length: '
                lengths=[ int(line.replace(length_string,'').split(' ')[0]) for line in err.splitlines() if length_string in line]
               
                if 200 in error_codes and max(lengths)>0:
                    success=True
            if not success:
                time.sleep(3*(trial+1))
        return success

    def download(self,var,pointer_var,checksum='',checksum_type='MD5',out_dir='.',version='v1'):
        dest_name=destination_download_files(self.url_name,out_dir,var,version,pointer_var)

        if checksum=='':
            if os.path.isfile(dest_name):
                return 'File '+dest_name+' found but could NOT check checksum of existing file because checksum was not a priori available. Not retrieving.'
        else:
            try: #Works only if file exists!
                comp_checksum=checksum_for_file(checksum_type,dest_name)
            except:
                comp_checksum=''
            if comp_checksum==checksum:
                return 'File '+dest_name+' found. '+checksum_type+' OK! Not retrieving.'

        with esgf_http.Dataset(self.url_name,
                                cache=self.cache,
                                timeout=self.timeout,
                                expire_after=self.expire_after,
                                session=self.session) as dataset:
            try: 
                file_size = int(dataset.response.headers['Content-Length'])
            except KeyError:
                file_size = False

            if file_size:
                size_string="Downloading: %s MB: %s" % (dest_name, file_size/2.0**20)
            else:
                size_string="Downloading: %s MB: Unknown" % (dest_name)
            
            directory=os.path.dirname(dest_name)
            if not os.path.exists(directory):
                os.makedirs(directory)
            file_size_dl = 0
            block_sz = 8192
            
            with open(dest_name, 'wb') as dest_file:
                for buffer in dataset.response.iter_content(block_sz):
                    file_size_dl += len(buffer)
                    dest_file.write(buffer)
                    if file_size:
                        status = r"%10d  [%3.2f%%]" % (file_size_dl, file_size_dl * 100. / file_size)
                        status = status + chr(8)*(len(status)+1)

        if checksum=='':
            return size_string+'\n'+'Could NOT check checksum of retrieved file because checksum was not a priori available.'
        else:
            try: 
                comp_checksum=checksum_for_file(checksum_type,dest_name)
            except:
                comp_checksum=''
            if comp_checksum!=checksum:
                try:
                    os.remove(dest_name)
                except:
                    pass
                return size_string+'\n'+'File '+dest_name+' does not have the same '+checksum_type+' checksum as published on the ESGF. Removing this file...'
            else:
                return size_string+'\n'+'Checking '+checksum_type+' checksum of retrieved file... '+checksum_type+' OK!'

def destination_download_files(url_name,out_dir,var,version,pointer_var):
    dest_name=out_dir.replace('tree','/'.join(pointer_var[:-1]))+'/'
    dest_name=dest_name.replace('var',var)
    dest_name=dest_name.replace('version',version)

    dest_name+=url_name.split('/')[-1]
    return os.path.abspath(os.path.expanduser(os.path.expandvars(dest_name)))

def md5_for_file(f, block_size=2**20):
    md5 = hashlib.md5()
    while True:
        data = f.read(block_size)
        if not data:
            break
        md5.update(data)
    return md5.hexdigest()

def sha_for_file(f, block_size=2**20):
    sha = hashlib.sha256()
    while True:
        data = f.read(block_size)
        if not data:
            break
        sha.update(data)
    return sha.hexdigest()

def checksum_for_file(checksum_type,dest_name, block_size=2**20):
    checksum = getattr(hashlib,checksum_type.lower())()
    with open(dest_name,'rb') as f:
        while True:
            data = f.read(block_size)
            if not data:
                break
            checksum.update(data)
    return checksum.hexdigest()



#External:
import netCDF4
import time
import os
import datetime
from socket import error as SocketError
import requests
import copy

#Internal:
import safe_handling
import esgf_pydap
import netcdf_utils

class queryable_netCDF:
    def __init__(self,file_name,semaphores=dict(),time_var='time',remote_data_node='',cache=None,timeout=120,expire_after=datetime.timedelta(hours=1),session=None):
        self.file_name=file_name
        self.semaphores=semaphores
        self.time_var=time_var

        if (remote_data_node in  self.semaphores.keys()):
            self.semaphore=semaphores[remote_data_node]
            self.handle_safely=True
        else:
            self.semaphore=safe_handling.dummy_semaphore()
            self.handle_safely=False

        self.cache=cache
        self.timeout=timeout
        self.expire_after=expire_after
        self.session=session
        if len(self.file_name)>4 and self.file_name[:4]=='http':
            self.use_pydap=True
            self.max_request=450
        else:
            self.use_pydap=False
            self.max_request=2048
        return

    def __enter__(self):
        self.semaphore.acquire()
        return self

    def __exit__(self,type,value,traceback):
        if self.handle_safely:
            #Do not release semaphore right away if data is not local:
            time.sleep(0.01)
        self.semaphore.release()
        return

    def unsafe_handling(self,function_handle,*args,**kwargs):
        try:
            #Capture errors. Important to prevent curl errors from being printed:
            redirection=safe_handling.suppress_stdout_stderr()
            if self.use_pydap:
                with esgf_pydap.Dataset(self.file_name,cache=self.cache,
                                        timeout=self.timeout,
                                        expire_after=self.expire_after,
                                        session=self.session) as dataset:
                    output=function_handle(dataset,*args,**kwargs)
            else:
                with redirection:
                    with netCDF4.Dataset(self.file_name) as dataset:
                        output=function_handle(dataset,*args,**kwargs)
        finally:
            redirection.close()
        return output

    def safe_handling(self,function_handle,*args,**kwargs):
        error_statement=' '.join('''
The url {0} could not be opened. 
Copy and paste this url in a browser and try downloading the file.
If it works, you can stop the download and retry using cdb_query. If
it still does not work it is likely that your certificates are either
not available or out of date.'''.splitlines()).format(self.file_name.replace('dodsC','fileServer'))
        num_trials=5
        success=False
        timeout=copy.copy(self.timeout)
        for trial in range(num_trials):
            if not success:
                try:
                    #Capture errors. Important to prevent curl errors from being printed:
                    if self.use_pydap:
                        with esgf_pydap.Dataset(self.file_name,
                                                cache=self.cache,
                                                timeout=timeout,
                                                expire_after=self.expire_after,
                                                session=self.session) as dataset:
                            output=function_handle(dataset,*args,**kwargs)
                    else:
                        try:
                            redirection=safe_handling.suppress_stdout_stderr()
                            with redirection:
                                with netCDF4.Dataset(self.file_name) as dataset:
                                    output=function_handle(dataset,*args,**kwargs)
                        finally:
                            redirection.close()
                    success=True
                except RuntimeError:
                    time.sleep(3*(trial+1))
                    pass
                except requests.exceptions.ReadTimeout as e:
                    time.sleep(3*(trial+1))
                    #Increase timeout:
                    timeout+=self.timeout
                    pass
                except requests.exceptions.ConnectionError as e:
                    time.sleep(3*(trial+1))
                    pass
                except SocketError as e:
                    #http://stackoverflow.com/questions/20568216/python-handling-socket-error-errno-104-connection-reset-by-peer
                    if e.errno != errno.ECONNRESET:
                        raise
                    time.sleep(3*(trial+1))
                    pass
        if not success:
            raise dodsError(error_statement)
        return output

    def check_if_opens(self,num_trials=5):
        error_statement=' '.join('''
The url {0} could not be opened. 
Copy and paste this url in a browser and try downloading the file.
If it works, you can stop the download and retry using cdb_query. If
it still does not work it is likely that your certificates are either
not available or out of date.'''.splitlines()).format(self.file_name.replace('dodsC','fileServer'))
        success=False
        for trial in range(num_trials):
            if not success:
                try:
                    #Capture errors. Important to prevent curl errors from being printed:
                    if self.use_pydap:
                        with esgf_pydap.Dataset(self.file_name,
                                                    cache=self.cache,
                                                    timeout=self.timeout,
                                                    expire_after=self.expire_after,
                                                    session=self.session) as dataset:
                            pass
                    else:
                        try:
                            redirection=safe_handling.suppress_stdout_stderr()
                            with redirection:
                                with netCDF4.Dataset(self.file_name) as dataset:
                                    pass
                        finally:
                            redirection.close()
                    success=True
                except RuntimeError:
                    time.sleep(3*(trial+1))
                    #print('Could have had a DAP error')
                    pass
                except requests.exceptions.ReadTimeout as e:
                    time.sleep(3*(trial+1))
                    pass
                except requests.exceptions.ConnectionError as e:
                    time.sleep(3*(trial+1))
                    pass
                except SocketError as e:
                    #http://stackoverflow.com/questions/20568216/python-handling-socket-error-errno-104-connection-reset-by-peer
                    if e.errno != errno.ECONNRESET:
                        raise
                    time.sleep(3*(trial+1))
                    pass
        return success

    def download(self,var,pointer_var,dimensions=dict(),unsort_dimensions=dict(),sort_table=[],time_var='time'):
        retrieved_data=self.safe_handling(
                         netcdf_utils.retrieve_container,var,
                                                        dimensions,
                                                        unsort_dimensions,
                                                        sort_table,self.max_request,
                                                        time_var=self.time_var
                                        )
        return (retrieved_data, sort_table, pointer_var+[var])

class dodsError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)



"""
Based on pydap.util.http
Frederic Laliberte,2016
"""

import re
from urlparse import urlsplit, urlunsplit

import requests
import requests_cache
import warnings

import pydap.lib
import pydap.client
from pydap.model import BaseType, SequenceType
from pydap.exceptions import ServerError
from pydap.parsers.dds import DDSParser
from pydap.parsers.das import DASParser
from pydap.xdr import DapUnpacker
from pydap.lib import walk, combine_slices, fix_slice, parse_qs, fix_shn, encode_atom

import os
import datetime
import numpy as np

from collections import OrderedDict

#Internal:
import esgf_pydap_proxy
import requests_sessions

class Dataset:
    def __init__(self,url,cache=None,expire_after=datetime.timedelta(hours=1),timeout=120,session=None):
        self.url=url
        self.cache=cache
        self.expire_after=expire_after
        self.timeout=timeout
        self.passed_session=session
        self.parent=self
        if (isinstance(self.passed_session,requests.Session) or
            isinstance(self.passed_session,requests_cache.core.CachedSession)
            ):
            self.session=self.passed_session
        else:
            self.session=requests_sessions.create_single_session(cache=self.cache,expire_after=self.expire_after)

        for response in [self._ddx, self._ddsdas]:
            self.dataset = response()
            if self.dataset: break
        else:
            raise ClientError("Unable to open dataset.")

        # Remove any projections from the url, leaving selections.
        scheme, netloc, path, query, fragment = urlsplit(self.url)
        projection, selection = parse_qs(query)
        url = urlunsplit(
                (scheme, netloc, path, '&'.join(selection), fragment))

        # Set data to a Proxy object for BaseType and SequenceType. These
        # variables can then be sliced to retrieve the data on-the-fly.
        for var in walk(self.dataset, BaseType):
            var.data = esgf_pydap_proxy.ArrayProxy(var.id, url, var.shape,self._request)
        for var in walk(self.dataset, SequenceType):
            var.data = esgf_pydap_proxy.SequenceProxy(var.id, url,self._request)

        # Set server-side functions.
        self.dataset.functions = pydap.client.Functions(url)

        # Apply the corresponding slices.
        projection = fix_shn(projection, self.dataset)
        for var in projection:
            target = self.dataset
            while var:
                token, slice_ = var.pop(0)
                target = target[token]
                if slice_ and isinstance(target.data, VariableProxy):
                    shape = getattr(target, 'shape', (sys.maxint,))
                    target.data._slice = fix_slice(slice_, shape)

        self.dimensions=self._dimensions()
        self.variables=self._variables()

        self._is_open=True
        return

    def __enter__(self):
        return self

    def isopen(self):
        return self._is_open

    def groups(self):
        return dict()
    
    def ncattrs(self):
        return self.dataset.attributes['NC_GLOBAL'].keys()

    def getncattr(self,attr):
        return self.dataset.attributes['NC_GLOBAL'][attr]

    def _dimensions(self):
        if ('DODS_EXTRA' in self.dataset.attributes.keys() and
            'Unlimited_Dimension' in self.dataset.attributes['DODS_EXTRA']):
            unlimited_dims=[self.dataset.attributes['DODS_EXTRA']['Unlimited_Dimension'],]
        else:
            unlimited_dims=[]
        var_list=self.dataset.keys()
        var_id=np.argmax(map(len,[self.dataset[varname].dimensions for varname in var_list]))
        base_dimensions_list=self.dataset[var_list[var_id]].dimensions
        base_dimensions_lengths=self.dataset[var_list[var_id]].shape
        
        for varname in var_list:
            if not set(base_dimensions_list).issuperset(self.dataset[varname].dimensions):
                for dim_id,dim in enumerate(self.dataset[varname].dimensions):
                    if not dim in base_dimensions_list:
                        base_dimensions_list+=(dim,)
                        base_dimensions_lengths+=(self.dataset[varname].shape[dim_id],)
        dimensions_dict=OrderedDict()
        for dim,dim_length in zip( base_dimensions_list,base_dimensions_lengths):
            dimensions_dict[dim]=Dimension(dim,dim_length,(dim in unlimited_dims),self.dataset)
        return  dimensions_dict

    def _variables(self):
        return {var:Variable(self.dataset[var],var,self.dataset) for var in self.dataset.keys()}

    def get_variables_by_attributes(self,**kwargs):
        #From netcdf4-python
        vs = []

        has_value_flag  = False
        for vname in self.variables:
            var = self.variables[vname]
            for k, v in kwargs.items():
                if callable(v):
                    has_value_flag = v(getattr(var, k, None))
                    if has_value_flag is False:
                        break
                #elif hasattr(var, k) and getattr(var, k) == v:
                #Must use getncattr
                elif hasattr(var, k) and var.getncattr(k) == v:
                    has_value_flag = True
                else:
                    has_value_flag = False
                    break
            if has_value_flag is True:
                vs.append(self.variables[vname])
        return vs

    def set_auto_mask(self,flag):
        raise NotImplementedError('set_auto_mask is not implemented for pydap')
        return

    def set_auto_scale(self,flag):
        raise NotImplementedError('set_auto_scale is not implemented for pydap')
        return

    def filepath(self):
        return self.url

    def _request(self,url):
        """
        Open a given URL and return headers and body.
        This function retrieves data from a given URL, returning the headers
        and the response body. Authentication can be set by adding the
        username and password to the URL; this will be sent as clear text
        only if the server only supports Basic authentication.
        """
        scheme, netloc, path, query, fragment = urlsplit(url)
        url = urlunsplit((
                scheme, netloc, path, query, fragment
                )).rstrip('?&')

        headers = {
            'user-agent': pydap.lib.USER_AGENT,
            'connection': 'close'}
        try:
            X509_PROXY=os.environ['X509_USER_PROXY']
        except KeyError:
            raise EnvironmentError('Environment variable X509_USER_PROXY must be set according to guidelines found at https://pythonhosted.org/cdb_query/install.html#obtaining-esgf-certificates')
            
        with warnings.catch_warnings():
             warnings.filterwarnings('ignore', message='Unverified HTTPS request is being made. Adding certificate verification is strongly advised. See: https://urllib3.readthedocs.org/en/latest/security.html')
             resp =self.session.get(url, 
                        cert=(X509_PROXY,X509_PROXY),
                        verify=False,
                        headers=headers,
                        allow_redirects=True,
                        timeout=self.timeout)

        # When an error is returned, we parse the error message from the
        # server and return it in a ``ClientError`` exception.
        try:
            if resp.headers["content-description"] in ["dods_error", "dods-error"]:
                m = re.search('code = (?P<code>[^;]+);\s*message = "(?P<msg>.*)"',
                        resp.content, re.DOTALL | re.MULTILINE)
                msg = 'Server error %(code)s: "%(msg)s"' % m.groupdict()
                raise ServerError(msg)
        finally:
            resp.raise_for_status()

        return resp.headers, resp.content

    def __exit__(self,type,value,traceback):
        self.close()
        return

    def close(self):
        if not (isinstance(self.passed_session,requests.Session) or
            isinstance(self.passed_session,requests_cache.core.CachedSession)
            ):
            #Close the session
            self.session.close()
        return
        
    def _ddx(self):
        """
        Stub function for DDX.

        Still waiting for the DDX spec to write this.

        """
        pass


    def _ddsdas(self):
        """
        Build the dataset from the DDS+DAS responses.

        This function builds the dataset object from the DDS and DAS
        responses, adding Proxy objects to the variables.

        """
        scheme, netloc, path, query, fragment = urlsplit(self.url)
        ddsurl = urlunsplit(
                (scheme, netloc, path + '.dds', query, fragment))
        dasurl = urlunsplit(
                (scheme, netloc, path + '.das', query, fragment))

        respdds, dds = self._request(ddsurl)
        respdas, das = self._request(dasurl)

        # Build the dataset structure and attributes.
        dataset = DDSParser(dds).parse()
        dataset = DASParser(das, dataset).parse()
        return dataset

_private_atts =\
['_grpid','_grp','_varid','groups','dimensions','variables','dtype','data_model','disk_format',
 '_nunlimdim','path','parent','ndim','mask','scale','cmptypes','vltypes','enumtypes','_isprimitive',
 'file_format','_isvlen','_isenum','_iscompound','_cmptype','_vltype','_enumtype','name',
 '__orthogoral_indexing__','keepweakref','_has_lsd']

class Variable:
    def __init__(self,var,name,dataset):
        self.var=var
        self.dimensions=self.var.dimensions
        if self.var.type.descriptor=='String':
            self.datatype='S1'
        else:
            self.datatype=self.var.type.typecode
        self.dtype=np.dtype(self.datatype)
        self.ndim=len(self.dimensions)
        self.shape=self.var.shape
        self.scale=True
        self.name=name
        self.size=np.prod(self.shape)
        self.dataset=dataset
        return

    def chunking(self):
        return 'contiguous'

    def filters(self):
        return None

    def get_var_chunk_cache(self):
        raise NotImplementedError('get_var_chunk_cache is not implemented')
        return

    #def __getattr__(self,name):
         # if name in _private_atts, it is stored at the python
        # level and not in the netCDF file.
        #if name.startswith('__') and name.endswith('__'):
        #    # if __dict__ requested, return a dict with netCDF attributes.
        #    if name == '__dict__':
        #        names = self.ncattrs()
        #        values = []
        #        for name in names:
        #            #values.append(_get_att(self._grp, self._varid, name))
        #            values.append(self.getncattr(name))
        #        return OrderedDict(zip(names,values))
        #    else:
        #        raise AttributeError
        #elif name in _private_atts:
        #    return self.__dict__[name]
        #else:
    #    return self.getncattr(name) 

    def ncattrs(self):
        return self.var.attributes.keys()

    def getncattr(self,attr):
        return self.var.attributes[attr]

    def getValue(self):
        return self.var[...]

    def group(self):
        return self.dataset

    #def __array__(self):
    #    return self[...]

    def __getitem__(self,getitem_tuple):
        try:
            return self.var.array.__getitem__(getitem_tuple)
        except (AttributeError, ServerError,requests.exceptions.HTTPError) as e:
            if ( 
                 isinstance(getitem_tuple,slice) and
                 getitem_tuple == phony_variable()[:]):
                #A single dimension ellipsis was requested. Use netCDF4 convention:
                return self[...]
            else:
                return self.var.__getitem__(getitem_tuple)

class phony_variable:
    #A phony variable to translate getitems:
    def __init__(self):
        pass

    def __getitem__(self,getitem_tuple):
        return getitem_tuple

class Dimension:
    def __init__(self,name,size,isunlimited,dataset):
        self.size=size
        self._isunlimited=isunlimited
        self.name=name
        self.dataset=dataset

    def __len__(self):
        return self.size

    def isunlimited(self):
        return self._isunlimited

    def group(self):
        return self.dataset






import numpy as np
import indices_utils

N=20
n=3
for id in range(10):
    indices=range(0,N,n)
    #indices=np.sort(np.random.choice(range(N),size=n,replace=False))

    slices=indices_utils.convert_indices_to_slices_fix(indices)
    print len(slices)
    print slices
    if np.any(np.sort(np.concatenate([np.arange(N)[simple_slice] for simple_slice in slices]))-indices)!=0:
        print indices




#External:
import os, sys, string, subprocess
import getpass
import select
import pydap.lib

#Internal:
from onlineca_get_trustroots_wget import onlineca_get_trustroots_wget
from onlineca_get_cert_wget import onlineca_get_cert_wget

def prompt_for_username_and_password(options):
    if (options.command=='certificates' and 
        'username' in dir(options) and options.username==None):
        options.username=raw_input('Enter Username: ')
        
    if 'username' in dir(options) and options.username!=None:
        if not options.password_from_pipe:
            options.password=getpass.getpass('Enter Credential phrase: ')
        else:
            prompt_timeout=1
            i,o,e=select.select([sys.stdin],[],[],prompt_timeout)
            if i:
                options.password=sys.stdin.readline()
            else:
                print '--password_from_pipe selected but no password was piped. Exiting.'
                return
    else:
        options.password=None

    #Retrieve certificates or set dods_conf:
    if 'username' in dir(options) and options.username!=None:
        retrieve_certificates(options.username,options.service,user_pass=options.password,trustroots=options.no_trustroots,timeout=options.timeout)
    else:
        retrieve_certificates(None,'ceda')
    return options

def retrieve_certificates(username,registering_service,user_pass=None,trustroots=False,timeout=120):
    home=os.getenv('HOME')

    dods=True
    dodsrc='%s/.dodsrc' % home
    esgfdir='%s/.esg4' % home
    if username==None and dods:
      #Set the environment:
      os.environ['DODS_CONF']=dodsrc
      #This modifies pydap. All three lines must be loaded in order:
      #pydap.lib.CACHE=esgfdir+'/dods_cache'
      #import pydap_esgf
      #pydap_esgf.install_esgf_client(os.environ['X509_USER_PROXY'],os.environ['X509_USER_PROXY'])
      return

    http_proxy=os.getenv('http_proxy')
    https_proxy=os.getenv('https_proxy')
    if http_proxy != None and https_proxy== None:
      print 'You have http_proxy set but not https_proxy: download tests are likely to fail'

    ee = { 'smhi':'esg-dn1.nsc.liu.se', 'pcmdi':'pcmdi9.llnl.gov', 'ipsl':'esgf-node.ipsl.fr', 'ceda':'slcs.ceda.ac.uk',
    'dkrz':'esgf-data.dkrz.de', 'pik':'esg.pik-potsdam.de', 'jpl':'jpl-esg.jpl.nasa.gov' }

    registering_service = ee.get( registering_service, registering_service )

    for f in [esgfdir,esgfdir+'/certificates',esgfdir+'/dods_cache']:
      if not os.path.isdir( f ):
        os.mkdir( f )

    dodstext = """## generated by enesGetCert
HTTP.COOKIEJAR=%(esgfdir)s/dods_cache/.dods_cookies
HTTP.SSL.CERTIFICATE=%(esgfdir)s/credentials.pem
HTTP.SSL.KEY=%(esgfdir)s/credentials.pem
HTTP.SSL.CAPATH=%(esgfdir)s/certificates
HTTP.TIMEOUT=%(timeout)s"""
#CURL.VERBOSE=0
#CURL.COOKIEJAR=.dods_cookies
#CURL.SSL.CERTIFICATE=%(esgfdir)s/credentials.pem
#CURL.SSL.KEY=%(esgfdir)s/credentials.pem
#CURL.SSL.CAPATH=%(esgfdir)s/certificates

    if dods:
      oo = open( dodsrc, 'w' )
      oo.write( dodstext % locals() )
      oo.close()
      #Temporary fix: write a local .dodsrc
      #oo = open( '.dodsrc', 'w' )
      #oo.write( dodstext % locals() )
      #oo.close()


    oo = open(esgfdir+'/onlineca-get-trustroots-wget.sh','w')
    oo.write(onlineca_get_trustroots_wget())
    oo.close()

    if trustroots:
        #Do not have to renew frequently
        call_to_script=['bash',esgfdir+'/onlineca-get-trustroots-wget.sh','-b','-U','https://'+registering_service+'/onlineca/trustroots/','-c',esgfdir+'/certificates']
        #print ' '.join(call_to_script)
        subprocess.call(call_to_script)

    oo = open(esgfdir+'/onlineca-cert-wget.sh','w')
    oo.write(onlineca_get_cert_wget())
    oo.close()
    call_to_script=['bash',esgfdir+'/onlineca-cert-wget.sh','-l',username,'-U','https://'+registering_service+'/onlineca/certificate/','-c',esgfdir]
    #print ' '.join(call_to_script)
    if user_pass!=None:
        call_to_script.append('-S')
        p=subprocess.Popen(call_to_script,stdout=subprocess.PIPE,stdin=subprocess.PIPE)
        stdout=p.communicate(input=user_pass)[0]
        p.stdin.close()
    else:
        subprocess.call(call_to_script)

    #Remove cookies for a clean slate:
    try:
        os.remove(esgfdir+'/dods_cache/.dods_cookies')
    except:
        pass
    return
    #port=MyProxyClient.PROPERTY_DEFAULTS['port']
    #lifetime=MyProxyClient.PROPERTY_DEFAULTS['proxyCertLifetime']
    #
    #if MyProxyClient.X509_CERT_DIR_ENVVARNAME in os.environ:
    #    cadir = os.environ[MyProxyClient.X509_CERT_DIR_ENVVARNAME]
    #else:
    #    cadir = os.path.join(
    #                    os.path.expanduser(MyProxyClient.USER_TRUSTROOT_DIR))
    #
    #client_props = dict(caCertDir=cadir,
    #                    hostname=registering_service,
    #                    port=port,
    #                    proxyCertLifetime=lifetime,
    #                    )
    #
    #myproxy = MyProxyClient(**client_props)
    #
    #creds = myproxy.logon(username, password,
    #                      bootstrap=True,
    #                      updateTrustRoots=False)
    #return

def test_certificates():
    home=os.getenv('HOME')
    http_proxy=os.getenv('http_proxy')
    https_proxy=os.getenv('https_proxy')
    if http_proxy != None and https_proxy== None:
      print 'You have http_proxy set but not https_proxy: download tests are likely to fail'

    esgfdir='%s/.esg4' % home

    print 'Testing certificate by running wget request in spider mode'
    cmd='wget -c -nH --certificate=%(esgfdir)s/credentials.pem --private-key=%(esgfdir)s/credentials.pem --save-cookies=%(esgfdir)s/cookies --load-cookies=%(esgfdir)s/cookies --ca-directory=%(esgfdir)s/certificates --no-check-certificate --spider   http://vesg.ipsl.fr/thredds/fileServer/esg_dataroot/CMIP5/output1/IPSL/IPSL-CM5A-LR/rcp85/mon/atmos/cfMon/r1i1p1/v20111119/clhcalipso/clhcalipso_cfMon_IPSL-CM5A-LR_rcp85_r1i1p1_200601-230012.nc  1> .wgsp 2> .wgspe' % locals()
    cmd2=['wget', '-c', '-nH', '--certificate=%(esgfdir)s/credentials.pem' %locals(), '--private-key=%(esgfdir)s/credentials.pem' %locals(), '--save-cookies=%(esgfdir)s/cookies' %locals(), '--load-cookies=%(esgfdir)s/cookies' %locals(), '--ca-directory=%(esgfdir)s/certificates' %locals(), '--no-check-certificate', '--spider', 'http://vesg.ipsl.fr/thredds/fileServer/esg_dataroot/CMIP5/output1/IPSL/IPSL-CM5A-LR/rcp85/mon/atmos/cfMon/r1i1p1/v20111119/clhcalipso/clhcalipso_cfMon_IPSL-CM5A-LR_rcp85_r1i1p1_200601-230012.nc' ] 
    if http_proxy != None:
      cmd = ( 'export http_proxy=%s ; export https_proxy=%s ;' % (http_proxy, https_proxy) ) + cmd
    print cmd
    subprocess.Popen( cmd2, env=os.environ.copy(), stdout=open('.wgsp','w'), stderr=open('.wgspe','w')  ).communicate()
    ##subprocess.Popen( cmd ).readlines()
    ii = open( '.wgspe' ).readlines()
    assert string.strip( ii[-2] ) in ['Remote file exists.','200 OK'] , 'File not found -- check .wgspe for error messages'
    os.unlink( '.wgsp' )
    os.unlink( '.wgspe' )

    print 'Check 1 OK'
    print '--------------------------------------------------'

    os.popen( 'ncdump -v 1> .nctmp 2> .nc2' ).readlines()
    ii = open( '.nc2' ).readlines()
    os.unlink( '.nctmp' )
    os.unlink( '.nc2' )
    try:
      x = string.split( string.split( ii[-1] )[3], '.' )
      maj = int(x[0])
      min = int(x[1])
      ncd = maj == 4 and min >= 1 or maj > 4
      if not ncd:
        print 'Netcdf libraries do not support opendap -- check 2 will not be completed'
    except:
      print 'Failed to identify ncdump version'
      ncd = False

    if ncd:
      print 'Testing certificate by requesting header of pr_day_HadGEM2-ES_esmControl_r1i1p1_20891201-20991130.nc'
      os.popen( 'ncdump -h http://cmip-dn1.badc.rl.ac.uk/thredds/dodsC/esg_dataroot/cmip5/output1/MOHC/HadGEM2-ES/esmControl/day/atmos/day/r1i1p1/v20120423/pr/pr_day_HadGEM2-ES_esmControl_r1i1p1_20891201-20991130.nc > .tmp ; md5sum .tmp > .md5' ).readlines()
      ii = open( '.md5' ).readlines()
      assert string.split(ii[0])[0] == '8f1d9ede885a527bcbdd1ddf1a5ed699', 'Checksum of header does not match expected value -- check .tmp and .md5 for clues'

      print 'Check 2 OK'
      print '--------------------------------------------------'

      print 'try: ncview http://cmip-dn1.badc.rl.ac.uk/thredds/dodsC/esg_dataroot/cmip5/output1/MOHC/HadGEM2-ES/esmControl/day/atmos/day/r1i1p1/v20120423/pr/pr_day_HadGEM2-ES_esmControl_r1i1p1_20891201-20991130.nc'
